with 
	AVR, 
	AVR.MCU, 
	AVR.Wait,
	AVR.Interrupts,
	Interfaces,
	interrupt_handler,
	System.Machine_Code,
	i_wire;


use AVR, AVR.MCU,Interfaces, interrupt_handler;

procedure Main is 

	current_limit: constant Unsigned_16 := Unsigned_16(4.0*(1024*0.185/5)*filter_pointer_type'modulus);
	night_limit: constant Unsigned_16 := Unsigned_16(0.3*(1024*0.185/5)*filter_pointer_type'modulus);
	search_band: constant Unsigned_8 := 20;
	
	PWM_A: constant Unsigned_8 := 2#10_00_00_11#;
				
	procedure Wait_ms is 
	new AVR.Wait.Generic_Busy_Wait_Milliseconds(Crystal_hertz => 9_600_000);
	
	
	procedure blink(h,l: Unsigned_16) is
	begin
		led := high;
		wait_ms(l);
		led := low;
		wait_ms(h);
	end blink;
	
	
	procedure Watchdog_reset is
	begin
		System.Machine_Code.Asm ("wdr", Volatile => True);
	end Watchdog_reset;	
	pragma Inline(Watchdog_reset);
	
	
	procedure Watchdog_disable is
	begin
		AVR.Interrupts.Disable;
		Watchdog_Reset;
		MCUSR_Bits(WDRF_Bit) := false;
		WDTCR_Bits(WDCE_Bit) := true;
		WDTCR_Bits(WDE_Bit) := true;
		WDTCR := 0;
		AVR.Interrupts.Enable;
	end Watchdog_disable;
	
	
	procedure Watchdog_enable is
	begin
		AVR.Interrupts.Disable;
		Watchdog_reset;
		WDTCR_Bits(WDCE_Bit) := true;
		WDTCR_Bits := (	WDTIF_Bit => true,
				WDTIE_Bit => true,
				WDP3_Bit => false,
				WDCE_Bit => false,
				WDE_Bit => false,			
				WDP2_Bit => false, WDP1_Bit => false, WDP0_Bit => false,
				others => false);
		AVR.Interrupts.Enable;
	end watchdog_enable;
	
	
	procedure Power_down is
	begin
		MCUCR_Bits(SM1_Bit) := true;
		MCUCR_Bits(SM0_Bit) := false;
		MCUCR_Bits(SE_bit) := true;
		System.Machine_Code.Asm ("sleep", Volatile => True);
		MCUCR_Bits(SE_bit) := false;
	end Power_Down;
	pragma Inline(Power_Down);
	
	
	procedure Set_Duty(d: Unsigned_8) is
	begin
		if not overcurrent_flag then OCR0A := d; end if;
		if overcurrent_flag then 
			OCR0A := 0;
			TCCR0A := 0;
			gate := low;
			loop blink(300,300); end loop;
		end if;
	end Set_Duty;
	
	
	procedure Step_Up is 
	begin
		if OCR0A<255 then Set_Duty(OCR0A+1); end if;
	end Step_Up;
	
	
	procedure Step_Down is 
	begin
		if OCR0A>0 then Set_Duty(OCR0A-1); end if;
	end Step_Down;
	
	
	function average_current return Unsigned_16 is
		ac: Unsigned_16 := 0;
	begin
		for i in current_filter'Range loop			
			ac := ac + Unsigned_16(current_filter(i));
		end loop;
		i_wire.Put(Unsigned_8(ac/filter_pointer_type'modulus));	
		return ac;		
	end;

	
	procedure near_MPPT is
		max_current: Unsigned_16 := 0;
		mpp: Unsigned_8 := 0;
		ac: Unsigned_16;
		bottom_bound, top_bound, search_start: Unsigned_8 := 0;
	begin
		led := high;		
		overvoltage_flag := false;
						
		if OCR0A<search_band then search_start := 0; else search_start := OCR0A - search_band; end if;
		
		for i in search_start..255 loop
			Set_Duty(i);
			wait_ms(5);			
			ac := average_current;
			
			if ac>max_current then bottom_bound := i; end if;
			
			if ac>=max_current then 
				max_current := ac;
				top_bound := i;				
			end if;
			
			if not safe_voltage then overvoltage_flag := true; end if;
			
			exit when not safe_voltage;
			
			exit when ac>=current_limit;
					
			mpp := Unsigned_8((Unsigned_16(top_bound)+Unsigned_16(bottom_bound))/2);
		end loop;
		
		Set_Duty(mpp);
		led := low;
	end near_MPPT;
		
		
	procedure idle(ms: Unsigned_16) is
	begin
		for i in 1..ms/4  loop	
			wait_ms(2);
			if not safe_voltage then overvoltage_flag := true; end if;
			
			if (average_current>current_limit) or (not safe_voltage)				
			then 
				Step_Down; 
			end if;

			if overcurrent_flag then Set_Duty(0); end if;			
		end loop;
		
	end idle;

	
	procedure Night is
	begin		
		while OCR0A<255 loop
			if not safe_voltage or average_current>night_limit then return; end if;
			Step_Up;
			wait_ms(1);			
		end loop;
		
		TCCR0A := 0;
		gate := high;
		Watchdog_enable;
		loop				
			Power_Down;
			exit when not safe_voltage;
			wait_ms(1);
			exit when average_current>night_limit;
		end loop;
		
		Watchdog_Disable;
		OCR0A := 0;
		TCCR0A := PWM_A;		
	end Night;

	
begin

	Watchdog_disable;
	
	
	LED := low;		
	DDRB_Bits(1) := DD_Output;
	
	debug_pin := low;
	DDRB_Bits(2) := DD_Output;
	
	
	-- 
	DDRB_Bits(4) := DD_Input;
	PORTB_Bits(4) := high;
	
	
	--  
	DDRB_Bits(3) := DD_Input;
	PORTB_Bits(3) := low;
	DIDR0_Bits := (ADC1D_Bit => true, others => false);
	ADMUX_Bits := (	REFS0_Bit=>false, 
			ADLAR_Bit=>false, 
			MUX1_Bit=>true, MUX0_Bit=>true, 
			others=>false);
	ADCSRA_Bits := (ADEN_Bit=>true, 
			ADSC_Bit=>true, 
			ADATE_Bit=>true, 
			ADIF_Bit=>false, 
			ADIE_BIt=>true, 
			ADPS2_Bit=>true, ADPS1_Bit=>false, ADPS0_Bit=>true);
	ADCSRB_Bits := (ADTS2_Bit=>false, ADTS1_Bit=>false, ADTS0_Bit=>false, others=>false);
	
	
	-- 
	PORTB_Bits(0) := low;
	DDRB_Bits(0) := DD_output;
	TCCR0A := PWM_A;
	TCCR0B := 2#00_00_0_001#; 
	OCR0A := 0;
	TCNT0 := 0;


	AVR.Interrupts.Enable;	
	
	
	loop 
		near_MPPT;		
		idle(10_000);		
		if (average_current<night_limit) and  safe_voltage and (not overvoltage_flag) then night; end if;
	end loop; 
	
end Main;
